from inspect import isclass
from shiftevent.event import Event, EventSchema
from shiftevent import exceptions as x
from shiftevent.default_handlers import default_handlers
from shiftevent.handlers import BaseHandler


class EventService:
    """
    Event service
    Responsible for handling events
    """

    # database instance
    db = None

    # event handlers
    handlers = None

    def __init__(self, db, handlers=None):
        """
        Initialize event service
        Accepts a database instance to operate on events and projections.
        :param db:
        """
        self.db = db
        self.handlers = handlers if handlers else default_handlers

    def event(
        self,
        type,
        object_id,
        author,
        payload=None,
        payload_rollback=None):
        """
        Persist an event
        Creates a new event object, validates it and saves to the database.
        May throw a validation exception if some event data is invalid.

        :param type: str, event type
        :param object_id: str, an id of the object being affected
        :param author:  str, author id in external system
        :param payload: dict, event payload
        :param payload_rollback: dict, payload to roll back an event
        :return: shiftevent.event.Event
        """
        # create
        event = Event(
            type=type,
            author=author,
            object_id=object_id,
            payload=payload,
            payload_rollback=payload_rollback
        )

        # check handler presence
        if type not in self.handlers:
            msg = 'No handlers for event of type [{}]'
            raise x.EventError(msg.format(type))

        # validate
        schema = EventSchema()
        ok = schema.process(event)
        if not ok:
            raise x.InvalidEvent(validation_errors=ok.get_messages())

        # and save
        events = self.db.tables['events']
        with self.db.engine.begin() as conn:
            data = event.to_db()
            del data['id']
            result = conn.execute(events.insert(), **data)
            event.id = result.inserted_primary_key[0]

        return event

    def emit(self, event):
        """
        Emit event
        Initialises every handler in the chain for the event and sequentially
        executes each one.
        :param event: shiftevent.events.event.Event
        :return:
        """
        if event.type not in self.handlers:
            raise x.EventError('No handlers for event {}'.format(event.type))

        # trigger handlers
        handlers = self.handlers[event.type]
        chain = []
        for handler in handlers:
            if not isclass(handler):
                msg = 'Handler {} for {} has to be a class, got [{}]'
                raise x.HandlerInstantiationError(msg.format(
                    handler,
                    event.type,
                    type(handler)
                ))

            handler = handler(db=self.db)
            if not isinstance(handler, BaseHandler):
                msg = 'Handler implementations must extend BaseHandler'
                raise x.HandlerInstantiationError(msg)

            # append to chain if valid
            chain.append(handler)

        # all valid? run chain
        for handler in chain:
            handled = handler.handle_event(event)
            if handled:
                event = handled
            else:
                break  # skip next handler

        # return event at the end
        return event

    def get_event(self, id):
        """
        Get event
        Returns event found by unique id.
        :param id: int, event id
        :return: shiftevent.event.Event
        """
        event = None
        events = self.db.tables['events']
        with self.db.engine.begin() as conn:
            select = events.select().where(events.c.id == id)
            data = conn.execute(select).fetchone()
            if data:
                event = Event(**data)
        return event











from datetime import datetime
from shiftevent import exceptions as x
from shiftschema.schema import Schema
from shiftschema import validators
from shiftschema import filters
import json
import copy


class EventSchema(Schema):
    """
    Event schema
    Defines filters and validators for an event
    """
    def schema(self):
        self.add_property('created')
        self.created.add_validator(validators.Required(
            message='An event must have creation date'
        ))

        self.add_property('type')
        self.type.add_filter(filters.Strip())
        self.type.add_filter(filters.Uppercase())
        self.type.add_validator(validators.Required(
            message='An event must have a type'
        ))

        self.add_property('object_id')
        self.object_id.add_filter(filters.Strip())
        self.object_id.add_validator(validators.Required(
            message='An event must have an object id'
        ))

        self.add_property('author')
        self.author.add_filter(filters.Strip())
        self.author.add_validator(validators.Required(
            message='An event must have an author set'
        ))


class Event:
    """
    Event
    Represent single atomic operation.
    """
    # event props, initialized at instance level
    props = dict()

    def __init__(self, *_, **kwargs):
        """
        Instantiate event object
        Can optionally populate itself from kwargs

        :param _: args, ignored
        :param kwargs: dict, key-value pairs used to populate event
        """
        # init props
        self.props = dict(
            id=None,
            created=None,
            type=None,
            author=None,
            object_id=None,
            payload=None,
            payload_rollback=None,
        )

        self.from_dict(kwargs)
        if not self.props['created']:
            self.props['created'] = datetime.utcnow()

    def __repr__(self):
        """ Returns printable representation of an event """
        repr = '<Event id=[{}] created=[{}] type=[{}] object_id=[{}]' \
               ' author=[{}]>'
        return repr.format(
            self.id,
            self.created,
            self.type,
            self.object_id,
            self.author
        )

    def __getattr__(self, item):
        """ Overrides attribute access for getting props """
        if item in self.props:
            return self.props[item]
        return object.__getattribute__(self, item)

    def __setattr__(self, key, value):
        """ Overrides attribute access for setting props """
        if key == 'payload':
            self.set_payload(value)
        elif key == 'payload_rollback':
            self.set_payload_rollback(value)
        elif key in self.props:
            self.props[key] = value
            return self
        else:
            object.__setattr__(self, key, value)
        return self

    @property
    def payload_json(self):
        """
        Payload json
        Returns payload as a json string
        :return: str
        """
        payload = self.payload if self.payload else {}
        return json.dumps(payload, ensure_ascii=False)

    @property
    def payload_rollback_json(self):
        """
        Payload rollback json
        Returns rollback payload as a json string
        :return: str
        """
        payload = self.payload_rollback if self.payload_rollback else {}
        return json.dumps(payload, ensure_ascii=False)

    def set_payload(self, payload):
        """
        Set payload
        Accepts a dictionary and encodes it into a json string for persistence.
        Will raise an exception if payload is not a dictionary.
        :param payload: dict or json string
        :return:
        """
        if type(payload) is str:
            try:
                payload = json.loads(payload, encoding='utf-8')
            except json.JSONDecodeError:
                raise x.EventError('Failed to decode payload string')

        if payload and type(payload) is not dict:
            msg = 'Payload must be a dictionary, got {}'
            raise x.EventError(msg.format(type(payload)))
        self.props['payload'] = payload
        return self

    def set_payload_rollback(self, payload):
        """
        Set rollback payload
        Accepts a dictionary and encodes it into a json string for persistence.
        Will raise an exception if payload is not a dictionary.
        :param payload: dict or json string
        :return:
        """
        if type(payload) is str:
            try:
                payload = json.loads(payload, encoding='utf-8')
            except json.JSONDecodeError:
                raise x.EventError('Failed to decode payload string')

        if payload and type(payload) is not dict:
            msg = 'Payload must be a dictionary, got {}'
            raise x.EventError(msg.format(type(payload)))
        self.props['payload_rollback'] = payload
        return self

    def to_dict(self):
        """ Returns dictionary representation of the event """
        return copy.copy(self.props)

    def to_db(self):
        """
        To db
        Returns db representation of event. Same as to dict, but payload is
        stringified to json. Used for persistence.
        :return:
        """
        data = self.to_dict()
        data['payload'] = self.payload_json
        data['payload_rollback'] = self.payload_rollback_json
        return data

    def from_dict(self, data):
        """ Populates itself from a dictionary """
        for prop, val in data.items():
            if prop in self.props:
                setattr(self, prop, val)
        return self













import json
from collections import Mapping
from sqlalchemy import create_engine
from sqlalchemy import MetaData
from sqlalchemy import sql
from sqlalchemy import desc, asc
from shiftevent.db_tables import define_tables
from shiftevent import exceptions as x


class Db:
    db_url = None
    db_params = None
    tables = dict()
    _meta = None
    _engine = None

    def __init__(
        self,
        db_url=None,
        engine=None,
        meta=None,
        dialect=None,
        **db_params
    ):
        """
        Instantiates database object
        Accepts database URL to connect to the engine  and a dict of db engine
        params that will be passed to te engine. See sqlalchmy engine docs for
        possible params: http://docs.sqlalchemy.org/en/latest/core/engines.html

        Alternatively can accept a ready-made engine via engine parameter which
        is useful for integration into applications when we don't need to
        manage separate connection pools.

        Additionally accepts a custom metadata object. Pass this if you want
        to integration tables in already existing metadata catalogue
        of your application.

        :param db_url: str, database url
        :param engine: sqlachemy engine
        :param meta: metadata object to attach to, optional
        :param dialect: str, only required for mysql
        :param db_params: parameters for engine creation (if not passed in)
        """
        if not db_url and not engine:
            msg = 'Can\'t instantiate database:db_url or engine required'
            raise x.DatabaseError(msg)

        self.db_url = db_url
        self.db_params = db_params
        self._engine = engine
        self._meta = meta
        self.tables = define_tables(self.meta, dialect=dialect)

    @property
    def engine(self):
        """
        Core interface to the database. Maintains connection pool.
        :return: sqlalchemy.engine.base.Engine
        """
        if not self._engine:
            self._engine = create_engine(self.db_url, **self.db_params)
        return self._engine

    @property
    def meta(self):
        """
        Metadata
        A catalogue of tables and columns.
        :return: sqlalchemy.sql.schema.MetaData
        """
        if not self._meta:
            self._meta = MetaData(self.engine)
        return self._meta

















import abc
from shiftevent import exceptions as x


class BaseHandler(metaclass=abc.ABCMeta):
    """
    Base event handler
    This enforces all event handlers to have a common interface and ensures
    all handlers have access to preconfigured environment objects like
    the database, cache, elastic search etc
    """

    # define event type in your concrete implementation
    EVENT_TYPES = ()

    # database instance
    db = None

    def __init__(self, db):
        """
        Initializes the handler and gets all required service injected
        :param db: shiftevent.db.Db
        """
        self.db = db

        # check event type defined
        if not self.EVENT_TYPES:
            msg = 'Event types undefined for handler [{}]'
            raise x.MissingEventType(msg.format(self.__class__))

    def check(self, event):
        """
        Check
        A shorthand to check if handler can support the event type passed in
        to handle/rollback methods.
        :param event: shiftevent.event.Event
        :return: bool
        """
        if event.type not in self.EVENT_TYPES:
            msg = 'Event handler {} can\'t support events of this type ({})'
            raise x.UnsupportedEventType(msg.format(self.__class__, event.type))

        return True

    def handle_event(self, event):
        """
        Handle event
        Wraps around user-defined handle method to run a check for
        event type support before actual execution.

        :param event: shiftevent.event.Event
        :return: shiftevent.event.Event
        """
        if not event.id:
            msg = 'Unable to handle unsaved event {}'
            raise x.ProcessingUnsavedEvent(msg.format(event))

        self.check(event)
        return self.handle(event)

    def rollback_event(self, event):
        """
        Rollback event
        Wraps around user-defined rollback method to run a check for
        event type support before actual execution.

        :param event: shiftevent.event.Event
        :return: shiftevent.event.Event
        :return:
        """
        if not event.id:
            msg = 'Unable to roll back unsaved event {}'
            raise x.ProcessingUnsavedEvent(msg.format(event))

        self.check(event)
        return self.rollback(event)

    @abc.abstractmethod
    def handle(self, event):
        """
        Process an event
        This should be implemented in concrete handler. It will get triggered
        once an event is emitted.

        :param event: shiftevent.event.Event
        :return: shiftevent.event.Event
        """
        raise NotImplemented('Implement me in your concrete handler')

    @abc.abstractmethod
    def rollback(self, event):
        """
        Rollback an event
        This should be implemented in concrete handler. It will get triggered
        once a travel back in time is requested and we have to sequentially
        undo events effectively reverting the changes made.

        :param event: shiftevent.event.Event
        :return: shiftevent.event.Event
        """
        raise NotImplemented('Implement me in your concrete handler')
